package main

import (
	"crypto/tls"
	"errors"
	"fmt"
	"net"
	"net/http"
	"os"
	"strings"
	"time"

	"git.polynom.me/rio/internal/acme"
	"git.polynom.me/rio/internal/certificates"
	"git.polynom.me/rio/internal/dns"
	"git.polynom.me/rio/internal/pages"
	"git.polynom.me/rio/internal/repo"
	"git.polynom.me/rio/internal/server"

	"code.gitea.io/sdk/gitea"
	legodns "github.com/go-acme/lego/v4/providers/dns"
	log "github.com/sirupsen/logrus"
	"github.com/urfave/cli/v2"
)

func handleSubdomain(domain, cname, path, giteaUrl, defaultCsp string, giteaClient *repo.GiteaClient, w http.ResponseWriter) {
	username := ""
	if cname != "" {
		// If we are accessed via a CNAME, then CNAME contains our <user>.<pages domain> value.
		username = strings.Split(cname, ".")[0]
	} else {
		// If we are directly accessed, then domain contains our <user>.<pages domain> value.
		username = strings.Split(domain, ".")[0]
	}

	// Strip the leading /
	if path[:1] == "/" {
		path = path[1:]
	}

	repo, path, err := repo.RepoFromPath(
		username,
		domain,
		cname,
		path,
		giteaClient,
	)
	if err != nil {
		log.Errorf("Failed to get repo: %s", err)
		w.WriteHeader(404)
		return
	}

	pages.ServeFile(username, repo.Name, path, defaultCsp, giteaClient, w)
}

func Handler(pagesDomain, giteaUrl, defaultCsp string, giteaClient *repo.GiteaClient) http.HandlerFunc {
	return func(w http.ResponseWriter, req *http.Request) {
		w.Header().Set("Server", "rio")

		if strings.HasSuffix(req.Host, pagesDomain) {
			log.Debug("Domain can be directly handled")
			handleSubdomain(req.Host, "", req.URL.Path, giteaUrl, defaultCsp, giteaClient, w)
			return
		}

		cname, err := dns.LookupCNAME(req.Host)
		if err != nil {
			log.Warningf("CNAME request failed, we don't handle %s", req.Host)
			w.WriteHeader(400)
			return
		}
		log.Debugf("Got CNAME %s", cname)

		if strings.HasSuffix(cname, pagesDomain) {
			log.Debugf("%s is alias of %s and can be handled after a CNAME query", req.Host, cname)
			handleSubdomain(req.Host, cname, req.URL.Path, giteaUrl, defaultCsp, giteaClient, w)
			return
		}

		log.Errorf("Not handling %s", req.Host)
		w.WriteHeader(404)
	}
}

func runServer(ctx *cli.Context) error {
	giteaUrl := ctx.String("gitea-url")
	domain := ctx.String("pages-domain")
	certsFile := ctx.String("certs-file")
	acmeEmail := ctx.String("acme-email")
	acmeServer := ctx.String("acme-server")
	acmeFile := ctx.String("acme-file")
	acmeDnsProvider := ctx.String("acme-dns-provider")
	acmeDisable := ctx.Bool("acme-disable")
	defaultCsp := ctx.String("default-csp")

	// Init Logging
	if ctx.Bool("debug") {
		log.SetLevel(log.DebugLevel)
	} else {
		log.SetLevel(log.InfoLevel)
	}

	// Setup the Gitea stuff
	httpClient := http.Client{Timeout: 10 * time.Second}
	giteaApiClient, err := gitea.NewClient(
		giteaUrl,
		gitea.SetHTTPClient(&httpClient),
		gitea.SetToken(""),
		gitea.SetUserAgent("rio"),
	)
	if err != nil {
		return err
	}
	giteaClient := repo.NewGiteaClient(giteaUrl, giteaApiClient)

	// Listen on the port
	addr := ctx.String("listen-host") + ":" + ctx.String("listen-port")
	listener, err := net.Listen("tcp", addr)
	if err != nil {
		errMsg := fmt.Errorf("Failed to create listener: %v", err)
		fmt.Println(errMsg.Error())
		return err
	}

	if !acmeDisable {
		if acmeEmail == "" || acmeFile == "" || certsFile == "" || acmeDnsProvider == "" {
			return errors.New("The options acme-dns-provider, acme-file, acme-email, and certs-file are required")
		}

		cache, err := certificates.CertificateCacheFromFile(certsFile)
		if err != nil {
			log.Debugf("Generating cert")
			fallback, err := certificates.MakeFallbackCertificate(domain)
			if err != nil {
				log.Fatalf("Failed to generate fallback certificate: %v", err)
				return err
			}

			cache.FallbackCertificate = fallback
			cache.Certificates = make(map[string]certificates.CertificateWrapper)
			cache.FlushToDisk(certsFile)
			log.Debug("Certificate wrote to disk")
		} else {
			log.Debug("Certificate store read from disk")
		}

		// Create an ACME client, if we failed to load one
		acmeClient, err := acme.ClientFromFile(acmeFile, acmeServer)
		if err != nil {
			log.Warn("Failed to load ACME client data from disk. Generating new account")
			acmeClient, err = acme.GenerateNewAccount(acmeEmail, acmeFile, acmeServer)
			if err != nil {
				log.Fatalf("Failed to generate new ACME client: %v", err)
				return err
			}
			log.Info("ACME account registered")
		} else {
			log.Info("ACME client data read from disk")
		}

		// Set up the DNS01 challenge solver
		provider, err := legodns.NewDNSChallengeProviderByName(acmeDnsProvider)
		if err != nil {
			log.Fatalf("Failed to create DNS01 challenge provider: %v", err)
			return err
		}
		err = acmeClient.Challenge.SetDNS01Provider(provider)
		if err != nil {
			log.Fatalf("Failed to setup DNS01 challenge solver: %v", err)
			return err
		}

		tlsConfig := server.MakeTlsConfig(
			domain,
			certsFile,
			&cache,
			acmeClient,
			&giteaClient,
		)
		listener = tls.NewListener(listener, tlsConfig)
	}

	if err := http.Serve(listener, Handler(domain, giteaUrl, defaultCsp, &giteaClient)); err != nil {
		fmt.Printf("Listening failed")
		return err
	}

	return nil
}

func main() {
	app := &cli.App{
		Action: runServer,
		Flags: []cli.Flag{
			&cli.StringFlag{
				Name:     "gitea-url",
				Usage:    "The (HTTPS) URL to the serving Gitea instance",
				EnvVars:  []string{"GITEA_URL"},
				Required: true,
			},
			&cli.StringFlag{
				Name:    "listen-host",
				Usage:   "The host to listen on",
				EnvVars: []string{"HOST"},
				Value:   "127.0.0.1",
			},
			&cli.StringFlag{
				Name:    "listen-port",
				Usage:   "The port to listen on",
				EnvVars: []string{"PORT"},
				Value:   "8888",
			},
			&cli.StringFlag{
				Name:    "acme-dns-provider",
				Usage:   "The provider to use for DNS01 challenge solving",
				EnvVars: []string{"ACME_DNS_PROVIDER"},
				Value:   "",
			},
			&cli.StringFlag{
				Name:     "pages-domain",
				Usage:    "The domain on which the server is reachable",
				EnvVars:  []string{"PAGES_DOMAIN"},
				Required: true,
			},
			&cli.StringFlag{
				Name:    "certs-file",
				Usage:   "File to store certificates in",
				EnvVars: []string{"CERTS_FILE"},
				Value:   "",
			},
			&cli.StringFlag{
				Name:    "acme-file",
				Usage:   "File to store ACME configuration in",
				EnvVars: []string{"ACME_FILE"},
				Value:   "",
			},
			&cli.StringFlag{
				Name:    "acme-email",
				Usage:   "Email to use for an ACME account",
				EnvVars: []string{"ACME_EMAIL"},
				Value:   "",
			},
			&cli.StringFlag{
				Name:    "acme-server",
				Usage:   "CA Directory to use",
				EnvVars: []string{"ACME_SERVER"},
				Value:   "https://acme-staging-v02.api.letsencrypt.org/directory",
			},
			&cli.BoolFlag{
				Name:    "acme-disable",
				Usage:   "Whether to disable automatic ACME certificates",
				EnvVars: []string{"ACME_DISABLE"},
			},
			&cli.BoolFlag{
				Name:    "debug",
				Usage:   "Whether to enable debug logging",
				EnvVars: []string{"DEBUG_ENABLE"},
			},
			&cli.StringFlag{
				Name:    "default-csp",
				Usage:   "The default CSP to include when sending HTTP responses",
				Value:   "",
				EnvVars: []string{"DEFAULT_CSP"},
			},
		},
	}

	if err := app.Run(os.Args); err != nil {
		log.Fatalf("Failed to run app: %s", err)
	}
}
